# Plague Utils
# Plague mod by Alex Mullins

from CvPythonExtensions import *
import CvUtil
import PyHelpers
import pickle

gc = CyGlobalContext()

#g_mPlagues = {}
#g_iNumPlagues = 0

PLAGUESTATUS_NO = 0
PLAGUESTATUS_HAS = 1
PLAGUESTATUS_CARRIER = 2
PLAGUESTATUS_IMMUNE = 3

g_mPlagueNames = ['Does not Have', 'Has', 'Carrier of', 'Immune to']

MIN_IMMUNE_AGE = 7
CHANCE_IMMUNE = 17
CHANCE_SPREAD_TRADE = 17
CHANCE_SPREAD_CITY_TO_UNIT = 14
CHANCE_SPREAD_UNIT_TO_CITY = 23
CHANCE_KILL_UNIT = 15

# phases of protection against a plague (cumulative)
PROT_NONE = 0	# no protection at all
PROT_MINOR = 1	# minor protection against plague spread and deaths
PROT_MAJOR = 2	# major protection against plague spread and deaths
PROT_SPREAD = 3	# plague can no longer spread to this civs units or cities
PROT_DEATH = 4	# plague can no longer cause deaths
PROT_TOTAL = 5	# end of protection levels (not possible to attain)

# required knowledge to gain levels of protection
g_ProtLvlKnowledge = [0, 6, 13, 18, 27]
g_szProtLvl = ['no protection from', 'minor protection from', 'major protection from', 'immunity to spread of', 'total immunity to']

#chance per 1000, cumulative ie 5,10 = 0-5 mutate, 5-10 disappear
CHANCE_MUTATE = 5
CHANCE_DISAPPEAR = 10

# chance per turn per modified bad health of a plague being generated
CHANCE_SPAWN = 0.01

def AddPlague(iVirility):
	try:
		mTable = pickle.loads(gc.getGame().getScriptData())
	except:
		mTable = {'mPlagues': {}, 'iNumPlagues': 0, 'iMinTimeNextPlague': 0}
	iID = mTable['iNumPlagues']
	mTable['iNumPlagues'] = mTable['iNumPlagues'] + 1
	mTable['mPlagues'][iID] = iVirility
	gc.getGame().setScriptData(pickle.dumps(mTable))
	return iID
	
def GetPlagueVirility(iID):
	try:
		mTable = pickle.loads(gc.getGame().getScriptData())
	except:
		# probably an error
		return 0
	return mTable['mPlagues'][iID]
	
def GetMinTimeNextPlague():
	try:
		mTable = pickle.loads(gc.getGame().getScriptData())
	except:
		mTable = {'mPlagues': {}, 'iNumPlagues': 0, 'iMinTimeNextPlague': 0}
		gc.getGame().setScriptData(pickle.dumps(mTable))
	return mTable['iMinTimeNextPlague']
	
def ResetMinTimeNextPlague():
	try:
		mTable = pickle.loads(gc.getGame().getScriptData())
	except:
		mTable = {'mPlagues': {}, 'iNumPlagues': 0, 'iMinTimeNextPlague': 0}
	mTable['iMinTimeNextPlague'] = 20 + gc.getGame().getSorenRandNum(20, "Time Next Plague")
	gc.getGame().setScriptData(pickle.dumps(mTable))
	
def CanStartPlague():
	if (GetMinTimeNextPlague() == 0):
		return true
	else:
		return false
		
def SpawnPlague(iID, pCity):
	try:
		mPlagues = pickle.loads(pCity.getScriptData())
	except:
		mPlagues = {}
	mPlagues[iID] = {'status': PLAGUESTATUS_HAS, 'age': 0}
	pCity.setScriptData(pickle.dumps(mPlagues))
	
def UpdatePlagues(pCity):
	try:
		mPlagues = pickle.loads(pCity.getScriptData())
	except:
		mPlagues = {}
	# loop through current plagues
	for iID, mPlague in mPlagues.items():
		if (mPlague['status'] == PLAGUESTATUS_HAS):
			if (gc.getGame().getSorenRandNum(100, "Plague Kill Chance") < GetPlagueVirility(iID)):
				# kill pop
				iPop = pCity.getPopulation()
				iPop = iPop - 1
				if (iPop > 0):
					pPlayer = gc.getPlayer(pCity.getOwner())
					UpdateKnowledge(pPlayer, iID)
					if (CanKill(pPlayer, iID)):
						pCity.setPopulation(iPop)
						sdEcho("City %s has lost a citizen to Plague #%d" %(pCity.getName(), iID))
					# end if
				# end if
			# end if
			mPlague['age'] = mPlague['age'] + 1
			if (mPlague['age'] > MIN_IMMUNE_AGE):
				if (gc.getGame().getSorenRandNum(100, "Plague Immune Chance") < CHANCE_IMMUNE):
					mPlague['status'] = gc.getGame().getSorenRandNum(2, "Plague Immune Type") + PLAGUESTATUS_CARRIER
					sdEcho("City %s now %s Plague #%d" %(pCity.getName(), g_mPlagueNames[mPlague['status']], iID))
				# end if
			# end if
			# check for spread to units in the city
			UpdatePlagues_CityToUnits(pCity)
			UpdatePlagues_UnitsToCity(pCity)
		elif (mPlague['status'] == PLAGUESTATUS_CARRIER):
			iRand = gc.getGame().getSorenRandNum(1000, "Plague Mutate Chance")
			if (iRand < CHANCE_MUTATE):
				iVirility = 20 + gc.getGame().getSorenRandNum(15, "Plague Mutant Virility")
				iNewID = AddPlague(iVirility)
				# bypass our SpawnPlague function so as not to mess with the scriptdata I'm going to overwrite at the end of this function
				mPlagues[iID] = {'status': PLAGUESTATUS_HAS, 'age': 0}
				sdEcho("A new Plague #%d has mutated from Plague #%d in %s, virility(%d)" %(iNewID, iID, pCity.getName(), iVirility))
			elif (iRand < CHANCE_DISAPPEAR):
				mPlague['status'] = PLAGUESTATUS_IMMUNE
				sdEcho("City %s has become immune to the dormant Plague #%d" %(pCity.getName(), iID))
			# end if
		# end if
	# end for
	pCity.setScriptData(pickle.dumps(mPlagues))
	# check conditions for new plague
	CheckCreatePlague(pCity)
	
def CheckCreatePlague(pCity):
	iBadHealth = pCity.getBadHealth()
	iGoodHealth = pCity.getGoodHealth()
	if (iBadHealth * 2.75 > iGoodHealth):
		# Chance of a plague starting, times 1000 to make it at least 1.0
		iPlagueChance = (iBadHealth * 2.75 - iGoodHealth) * CHANCE_SPAWN * 1000
		# Should we start a new plague here?
		iRand = gc.getGame().getSorenRandNum(100000, "Plague Chance")
		if (iRand < iPlagueChance):
			# Start a plague in this city
			iVirility = (iBadHealth * 2.75 - iGoodHealth) * gc.getGame().getSorenRandNum(15, "Plague Virility")
			if (iVirility < 5):
				iVirility = 5
			# end if
			iID = AddPlague(iVirility)
			SpawnPlague(iID, pCity)
			sdEcho("A new Plague #%d has appeared in %s, virility(%d)" %(iID, pCity.getName(), iVirility))
		#end if
	#end if
	
def SpreadPlagues(pCity):
	try:
		mPlagues = pickle.loads(pCity.getScriptData())
	except:
		mPlagues = {}
	for iID, mPlague in mPlagues.items():
		if (mPlague['status'] > PLAGUESTATUS_NO and mPlague['status'] < PLAGUESTATUS_IMMUNE):
			# Loop through all of City's trade routes
			for iTargetCityCursor in range(pCity.GetCy().getTradeRoutes()):
				pTargetCity = pCity.GetCy().getTradeCity(iTargetCityCursor)
				if (not pTargetCity.isNone() and not CityHasPlague(iID, pTargetCity) and gc.getGame().getSorenRandNum(100, "Plague Spread Trade Chance") < CHANCE_SPREAD_TRADE):
					pTargetPlayer = gc.getPlayer(pTargetCity.getOwner())
					UpdateKnowledge(pTargetPlayer, iID)
					if (CanSpread(pTargetPlayer, iID)):
						SpawnPlague(iID, pTargetCity)
						sdEcho("Plague #%d has spread from %s to %s via a trade route" %(iID, pCity.getName(), pTargetCity.getName()))
					# end if
				# end if
			# end for
		#end if
	#end for
	
def UpdatePlagues_CityToUnits(pCity):
	try:
		mPlagues = pickle.loads(pCity.getScriptData())
	except:
		# no plagues to spread
		return
	pPlot = pCity.plot()
	for iID, mPlague in mPlagues.items():
		if (mPlague['status'] == PLAGUESTATUS_HAS or mPlague['status'] == PLAGUESTATUS_CARRIER):
			# loop through all the units in this plot
			for i in range(pPlot.getNumUnits()):
				pUnit = pPlot.getUnit(i)
				if (not UnitHasPlague(iID, pUnit)):
					if (gc.getGame().getSorenRandNum(100, "Plague Spread to Unit Chance") < CHANCE_SPREAD_CITY_TO_UNIT):
						# give the unit the plague
						# test to see if it kills the unit
						if (gc.getGame().getSorenRandNum(100, "Plague Kill Unit Chance") < CHANCE_KILL_UNIT):
							pPlayer = gc.getPlayer(pCity.getOwner())
							UpdateKnowledge(pPlayer, iID)
							if (CanKill(pPlayer, iID)):
								pUnit.kill(false, -1)
								sdEcho("Plague #%d has killed %s in %s!" %(iID, pUnit.getName(), pCity.getName()))
							# end if
						else:
							pPlayer = gc.getPlayer(pCity.getOwner())
							UpdateKnowledge(pPlayer, iID)
							if (CanSpread(pPlayer, iID)):
								iDamage = 100 - pUnit.getDamage()
								iDamage = int(iDamage * gc.getGame().getSorenRandNum(80, "Plague Unit Damage") * .01)
								iDamageTaken = iDamage - pUnit.getDamage()
								if (iDamageTaken < 0):
									iDamageTaken = 0
								# end if
								sdEcho("%s in %s has contracted Plague #%d and taken %d damage" %(pUnit.getName(), pCity.getName(), iID, iDamageTaken))
								pUnit.setDamage(iDamage, -1)
								GiveUnitPlague(iID, pUnit)
								pUnit.setHasPromotion(gc.getInfoTypeForString("PROMOTION_PLAGUE"), true)
							# end if
						# end if
					# end if
				# end if
			# end for
		# end if
	# end for
	
def UpdatePlagues_UnitsToCity(pCity):
	# Check if each unit transmits a plague to the city
	pPlot = pCity.plot()
	for i in range(pPlot.getNumUnits()):
		pUnit = pPlot.getUnit(i)
		try:
			mPlagues = pickle.loads(pUnit.getScriptData())
		except:
			# no plagues
			continue
		for iID, mPlague in mPlagues.items():
			if (not CityHasPlague(iID, pCity)):
				# check for spread
				if (gc.getGame().getSorenRandNum(100, "Plague Spread Unit to City Chance") < CHANCE_SPREAD_UNIT_TO_CITY):
					pPlayer = gc.getPlayer(pCity.getOwner())
					UpdateKnowledge(pPlayer, iID)
					if (CanSpread(pPlayer, iID)):
						sdEcho("Plague #%d has spread from %s to %s" %(iID, pUnit.getName(), pCity.getName()))
						SpawnPlague(iId, pCity)
					# end if
				# end if
			# end if
		# end for
	# end for
		
def UpdateUnitPlagues(pUnit):
	try:
		mPlagues = pickle.loads(pUnit.getScriptData())
	except:
		# this unit has no plagues to update
		return
	for iID, mPlague in mPlagues.items():
		if (mPlague['status'] == PLAGUESTATUS_HAS):
			mPlague['age'] = mPlague['age'] + 1
			if (mPlague['age'] > MIN_IMMUNE_AGE):
				if (gc.getGame().getSorenRandNum(100, "Plague Immune Chance") < CHANCE_IMMUNE):
					pPlayer = gc.getPlayer(pUnit.getOwner())
					mPlague['status'] = gc.getGame().getSorenRandNum(2, "Plague Immune Type") + PLAGUESTATUS_CARRIER
					sdEcho("%s's Unit %s now %s Plague #%d" %(pPlayer.getName(), pUnit.getName(), g_mPlagueNames[mPlague['status']], iID))
					pUnit.setHasPromotion(gc.getInfoTypeForString("PROMOTION_PLAGUE"), false)
					pUnit.setHasPromotion(gc.getInfoTypeForString("PROMOTION_PLAGUE_CARRIER"), true)
				# end if
			# end if
		elif (mPlague['status'] == PLAGUESTATUS_CARRIER):
			iRand = gc.getGame().getSorenRandNum(1000, "Plague Mutate Chance")
			# no mutating in units so subtract the mutate chance to get the true disappear chance
			if (iRand < CHANCE_DISAPPEAR - CHANCE_MUTATE):
				pPlayer = gc.getPlayer(punit.getOwner())
				mPlague['status'] = PLAGUESTATUS_IMMUNE
				sdEcho("%s's Unit %s has become immune to the dormant Plague #%d" %(pPlayer.getName(), pUnit.getName(), iID))
				pUnit.setHasPromotion(gc.getInfoTypeForString("PROMOTION_PLAGUE_CARRIER"), false)
			# end if
		# end if
	# end for
	pUnit.setScriptData(pickle.dumps(mPlagues))

def UnitHasPlague(iTestID, pUnit):
	try:
		mPlagues = pickle.loads(pUnit.getScriptData())
	except:
		mPlagues = {}
	for iID, mPlague in mPlagues.items():
		if (iID == iTestID):
			return true
		#end if
	#end for
	return false

def GiveUnitPlague(iID, pUnit):
	try:
		mPlagues = pickle.loads(pUnit.getScriptData())
	except:
		mPlagues = {}
	mPlagues[iID] = {'status': PLAGUESTATUS_HAS, 'age': 0}
	pUnit.setScriptData(pickle.dumps(mPlagues))
	
def CityHasPlague(iTestID, pCity):
	try:
		mPlagues = pickle.loads(pCity.getScriptData())
	except:
		mPlagues = {}
	for iID, mPlague in mPlagues.items():
		if (iID == iTestID):
			return true
		#end if
	#end for
	return false
	
# this should be called before checking if a plague can spread or kill
# succesfully preventing a spread/death gains knowledge
def UpdateKnowledge(pPlayer, iID):
	try:
		mTable = pickle.loads(pPlayer.getScriptData())
	except:
		mTable = {}
	if (not mTable.has_key(iID)):
		mTable[iID] = {'knowledge': 0, 'protection': PROT_NONE}
	if (gc.getGame().getSorenRandNum(100, "Plague Gain Knowledge Chance") < ChanceGainKnowledge(pPlayer)):
		mTable[iID]['knowledge'] = mTable[iID]['knowledge'] + 1
		sdEcho("%s's civilization has gained knowledge about Plague #%d" %(pPlayer.getName(), iID))
		if (mTable[iID]['knowledge'] > g_ProtLvlKnowledge[mTable[iID]['protection'] + 1] and mTable[iID]['protection'] + 1 < PROT_TOTAL):
			mTable[iID]['protection'] = mTable[iID]['protection'] + 1
			sdEcho("%s's civilization now has %s to Plague #%d" %(pPlayer.getName(), g_szProtLvl[mTable[iID]['protection']], iID))
		# end if
		pPlayer.setScriptData(pickle.dumps(mTable))
	# end if
		
# returns if a plague can spread
# checks against protections a player has
# should be called every time a plague attempts to spread
def CanSpread(pPlayer, iID):
	try:
		mTable = pickle.loads(pPlayer.getScriptData())
	except:
		mTable = {}
	if (not mTable.has_key(iID)):
		mTable[iID] = {'knowledge': 0, 'protection': PROT_NONE}
	# end if
	if (mTable[iID]['protection'] >= PROT_SPREAD):
		return false
	elif (mTable[iID]['protection'] >= PROT_MAJOR):
		if (gc.getGame().getSorenRandNum(100, "Plague Protection") < 66):
			return true
		else:
			return false
		# end if
	elif (mTable[iID]['protection'] >= PROT_MINOR):
		if (gc.getGame().getSorenRandNum(100, "Plague Protection") < 33):
			return true
		else:
			return false
		# end if
	else:
		return true
	# end if
	
# returns if a plague can kill
# checks against protections a player has
# should be called every time a plague attempts to spread
def CanKill(pPlayer, iID):
	try:
		mTable = pickle.loads(pPlayer.getScriptData())
	except:
		mTable = {}
	if (not mTable.has_key(iID)):
		mTable[iID] = {'knowledge': 0, 'protection': PROT_NONE}
	# end if
	if (mTable[iID]['protection'] >= PROT_DEATH):
		return false
	elif (mTable[iID]['protection'] >= PROT_SPREAD):
		if (gc.getGame().getSorenRandNum(100, "Plague Protection") < 75):
			return true
		else:
			return false
		# end if
	elif (mTable[iID]['protection'] >= PROT_MAJOR):
		if (gc.getGame().getSorenRandNum(100, "Plague Protection") < 50):
			return true
		else:
			return false
		# end if
	elif (mTable[iID]['protection'] >= PROT_MINOR):
		if (gc.getGame().getSorenRandNum(100, "Plague Protection") < 25):
			return true
		else:
			return false
		# end if
	else:
		return true
	# end if

# the chance a civilization will gain knowledge about a certain plague upon a death or spread of it within their civ
# this percentage chance should start low and increase as a player gains technology
def ChanceGainKnowledge(pPlayer):
	return 75

# "borrowed" from sdToolKit
def sdEcho( echoString ):
	printToScr = True
	printToLog = True
	message = "%s" %(echoString)
	if (printToScr):
		CyInterface().addImmediateMessage(message,"")
	if (printToLog):
		CvUtil.pyPrint(message)
	return 0